
(* Coq certification for conjectures about `add' from the Cyclist's distribution

Contact: sorin.stratulat@loria.fr

We use the certification methodology 
from our paper "Mechanically 
certifying formula-based {N}oetherian 
induction reasoning" 

*) 

(* import the syntactic ordering based on the Coccinelle library *)

Require Import ordering.

Section add.
(* T is the unique datatype. All the functions and inductive types are defined only on T. *)

Variable T: Set.
Variable zero: T.
Variable succ: T -> T.

Inductive N: T -> Prop :=
  n1 : N zero
| n2 : forall x, N x -> N (succ x).

Inductive Add:  T ->  T ->  T -> Prop :=
  add1: forall x, N x -> Add zero x x
| add2: forall x y z, Add x y z -> Add (succ x) y (succ z)
.

(* the translation function to Coccinelle terms *)

Variable model_nat : T -> term.
Axiom model_nat_0: model_nat zero = (Term id_0 nil).
Axiom  model_nat_succ: forall (u1: T), model_nat (succ u1) = (Term id_S ((model_nat u1)::nil)).

Hint Rewrite model_nat_0 model_nat_succ : model_nat.
Ltac rewrite_model := autorewrite with model_nat.


Section add_06.

(* Coq certification for 06-zero-plus-zero.tst.no. It does not need induction reasoning *)

(* prove the root conjectures *)

Theorem true_0: Add zero zero zero.
Proof.
  apply add1. apply n1.
Qed.

Reset true_0.
End add_06.

Section add_07.

(* Coq certification for 07-a-plus-zero.tst *)


(* Initialisation of the proof process *)

Definition type_LF_0 :=  T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_0 : type_LF_0:= (fun u1 => (N u1 -> Add u1 zero u1, ((model_nat u1):: nil))).

(* 2. build the list of Fs *)

Definition LF_0 := [F_0].

(* 3. build the main lemma *)
      
Lemma main_0 : forall F, In F LF_0 -> forall u1, (forall F', In F' LF_0 -> forall e1, less (snd (F' e1)) (snd (F u1)) -> fst (F' e1)) -> fst (F u1).
Proof.
intros F HF u1; case_In HF; intro Hind.
rename u1 into x.

simpl.

intros.

(* instantiate x *)
(* add(x,0,x) *) generalize Hind. inversion H; intros.

(* add(0,0,0) *) rewrite <- H0. apply add1.  apply n1.

clear H1 Hind x H. rename x0 into y.


(* add(s(y),0,s(y)) *) apply add2.


(* induction step for add(y,0,y) *)


(* add(y,0,y) *) apply (Hind0 F_0). 
simpl. left. trivial.
unfold snd. unfold F_0. rewrite_model. abstract solve_rpo_mul.

trivial. 
Qed.

Definition S_0 := fun f => exists F, In F LF_0 /\ exists e1, f = F e1.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_0: forall F, In F LF_0 -> forall u1: T, fst (F u1).
Proof.
  let n := constr:(2) in
let p := constr:(S(n)) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_0);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_0;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjecture *)
      
Theorem true_0: forall u1:T, N u1 -> Add u1 zero u1.
Proof.
  do 1 intro.
apply (all_true_0 F_0).
 (trivial_in 0) ||
                (repeat constructor).
Qed.

Reset type_LF_0.
End add_07.

Section add_08.


(* Coq certification for 08-sum-is-nat.tst *)


(* process the first strongly connected component *)

Local Definition type_LF_0 :=  T -> T -> T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_0 : type_LF_0:= (fun u1 u2 u3 => (N u1 -> N u2 -> Add u1 u2 u3 -> N u3, ((model_nat u1)::(model_nat u3):: nil))).

(* 2. build the list of Fs *)

Definition LF_0 := [F_0].

(* 3. build the main lemma *)
      
Axiom  zero_diff_succ: forall (u1: T), zero = (succ u1) -> False.
Axiom injective_succ: forall u1 u2, succ u1 = succ u2 -> u1 = u2.

Lemma main_0 : forall F, In F LF_0 -> forall u1 u2 u3, (forall F', In F' LF_0 -> forall e1 e2 e3, less (snd (F' e1 e2 e3)) (snd (F u1 u2 u3)) -> fst (F' e1 e2 e3)) -> fst (F u1 u2 u3).
Proof.
intros F HF x y z; case_In HF; intro Hind.


simpl. intros.

(* instantiate using Add *) 
(* N(z) *) generalize Hind. inversion H1; intros.

(* N(z) *) trivial.

rename z0 into b. 
(* N(s(b)) *) apply n2.

rename x0 into w. rewrite <- H3 in H1. rewrite <- H5 in H1. rewrite <- H3 in H. clear Hind H3 x H5 x z y0 H4.

(* instantiate y *) 

(* N(y) *) generalize Hind0. inversion H; intros.

(* Ex Falso *)

(* N(b) *) assert False. apply (zero_diff_succ w). trivial. destruct H4.

apply injective_succ in H3. rewrite H3 in H4.
clear Hind1 x H3.

(* induction step for N(b) *)


(* N(b) *) assert (H':= Hind0 F_0). unfold F_0 in H'. unfold fst in H'. apply H' with (e1:= w) (e2:=y).
simpl. left. trivial.
unfold snd. unfold F_0. rewrite_model. abstract solve_rpo_mul.

trivial. trivial.
trivial. 
Qed.




Definition S_0 := fun f => exists F, In F LF_0 /\ exists e1 e2 e3, f = F e1 e2 e3.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_0: forall F, In F LF_0 -> forall u1 u2 u3: T, fst (F u1 u2 u3).
Proof.
let n := constr:(4) in
let p := constr:(S(n)) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_0);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_0;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjecture*)

Theorem true_0: forall u1 u2 u3, N u1 -> N u2 -> Add u1 u2 u3 -> N u3.
Proof.
  do 2 intro.
apply (all_true_0 F_0);
 (trivial_in 0) ||
 (repeat constructor).
Qed.
Reset type_LF_0.
End add_08.


Section add_09.


(* Coq certification for 09-a-plus-b-plus-one.tst *)


(* process the first strongly connected component *)

Local Definition type_LF_0 :=  T -> T -> T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_0 : type_LF_0:= (fun u1 u2 u3 => (N u1 -> N u2 -> Add u1 u2 u3 -> Add u1 (succ u2) (succ u3), ((model_nat u1)::(model_nat u3):: nil))).

(* 2. build the list of Fs *)

Definition LF_0 := [F_0].

(* 3. build the main lemma *)
      
Axiom  zero_diff_succ: forall (u1: T), zero = (succ u1) -> False.
Axiom injective_succ: forall u1 u2, succ u1 = succ u2 -> u1 = u2.

Lemma main_0 : forall F, In F LF_0 -> forall u1 u2 u3, (forall F', In F' LF_0 -> forall e1 e2 e3, less (snd (F' e1 e2 e3)) (snd (F u1 u2 u3)) -> fst (F' e1 e2 e3)) -> fst (F u1 u2 u3).
Proof.
intros F HF x y z; case_In HF; intro Hind.


simpl. intros.

(* instantiate using Add *) 
(* Add x s(y) s(z) *) generalize Hind. inversion H; intros.

(* Add 0 s(y) s(z) *)  inversion H1. rewrite <- H4. apply add1.
apply n2. trivial.

rewrite <- H4 in H2. apply zero_diff_succ in H2. destruct H2.


rename x0 into w.
(* Add s(w) s(y) s(z) *) apply add2.

rewrite <- H3 in H1. 
clear x Hind H H3.

(* Add w s(y) z *) generalize Hind0. inversion H1; intros.

apply zero_diff_succ in H3. destruct H3.

rename z0 into c. apply injective_succ in H. rewrite H in H3. rewrite H in Hind1.

clear H Hind0 x y0 H4 H5.

(* induction step for Add w s(y) s(c) *)


(* Add w s(y) s(c) *) apply (Hind1 F_0). 
simpl. left. trivial.
unfold snd. unfold F_0. rewrite_model. abstract solve_rpo_mul.

trivial. trivial.
trivial. 
Qed.
End add_09.
End add.
