
(* Coq certification for conjectures about `even' and `odd' from the Cyclist's distribution

Contact: sorin.stratulat@loria.fr

We use the certification methodology 
from our paper "Mechanically 
certifying formula-based {N}oetherian 
induction reasoning" 

*) 

(* import the syntactic ordering based on the Coccinelle library *)

Require Import ordering.

Section evenodd.

(* T is the unique datatype. All the functions and inductive types are defined only on T. *)

Variable T: Set.
Variable zero: T.
Variable succ: T -> T.


Inductive E:  T -> Prop :=
  e1: E zero
| e2: forall x, O x -> E (succ x)
with
 O:  T -> Prop :=
 o1: forall x, E x -> O (succ x).

Inductive N: T -> Prop :=
  n1 : N zero
| n2 : forall x, N x -> N (succ x).


(* the translation function to Coccinelle terms *)

Variable model_nat : T -> term.
Axiom model_nat_0: model_nat zero = (Term id_0 nil).
Axiom  model_nat_succ: forall (u1: T), model_nat (succ u1) = (Term id_S ((model_nat u1)::nil)).

Hint Rewrite model_nat_0 model_nat_succ : model_nat.
Ltac rewrite_model := autorewrite with model_nat.

Section even_odd_01.

(* Coq certification for 01-odd-even.tst *)

(* Initialisation of the proof process *)

Definition type_LF_0 :=  T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_0 : type_LF_0:= (fun u1 => (O u1 -> N u1, ((model_nat u1):: nil))).

(* 2. build the list of Fs *)

Definition LF_0 := [F_0].

(* 3. build the main lemma *)
      
Lemma main_0 : forall F, In F LF_0 -> forall u1, (forall F', In F' LF_0 -> forall e1, less (snd (F' e1)) (snd (F u1)) -> fst (F' e1)) -> fst (F u1).
Proof.
intros F HF u1; case_In HF; intro Hind.
rename u1 into x.

simpl.

intros.

(* instantiate x *)
 generalize Hind.

(* N(x) *) inversion H. clear H1. intros. rename x0 into y.


(* N(s(y)) *) apply n2.

clear Hind H x.

(* instantiate y *)

generalize Hind0.

(* N(y) *) inversion H0.

(* N(0) *) intros. rewrite <- H. apply n1.

intros. rename x into z.
(* N(s(z)) *) apply n2.

clear H0 H1 Hind0.

(* induction step for N(z) *)


(* N(z) *) apply (Hind1 F_0). 
simpl. left. trivial.
unfold snd. unfold F_0. rewrite_model. abstract solve_rpo_mul.

trivial. 
Qed.

Definition S_0 := fun f => exists F, In F LF_0 /\ exists e1, f = F e1.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_0: forall F, In F LF_0 -> forall u1: T, fst (F u1).
Proof.
  let n := constr:(2) in
let p := constr:(S(n)) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_0);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_0;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjecture *)
      
Theorem true_0: forall x:T, O x -> N x.
Proof.
  do 1 intro.
apply (all_true_0 F_0).
 (trivial_in 0) ||
                (repeat constructor).
Qed.

Reset type_LF_0.

End even_odd_01.

Section even_odd_02.

(* process the first strongly connected component *)

Local Definition type_LF_1 :=  T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_1 : type_LF_1:= (fun u1 => (E u1 -> N u1, ((model_nat u1):: nil))).

(* 2. build the list of Fs *)

Definition LF_1 := [F_1].

(* 3. build the main lemma *)
      
Lemma main_1 : forall F, In F LF_1 -> forall u1, (forall F', In F' LF_1 -> forall e1, less (snd (F' e1)) (snd (F u1)) -> fst (F' e1)) -> fst (F u1).
Proof.
intros F HF u1; case_In HF; intro Hind.
rename u1 into x.

simpl. intros.

(* instantiate x *) 
(* N(x) *) generalize Hind. inversion H; intros.

(* N(0) *) rewrite <- H0. apply n1.

rename x0 into y. 
(* N(s(y) *) apply n2.

clear Hind H x H1.

(* instantiate y *) 

(* N(y) *) generalize Hind0. inversion_clear H0; intros.

rename x into z.

(* N(s(z)) *) apply n2. clear Hind0 y.

(* induction step for N(z) *)


(* N(z) *) apply (Hind1 F_1). 
simpl. left. trivial.
unfold snd. unfold F_1. rewrite_model. abstract solve_rpo_mul.

trivial. 
Qed.




Definition S_1 := fun f => exists F, In F LF_1 /\ exists e1, f = F e1.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_1: forall F, In F LF_1 -> forall u1: T, fst (F u1).
Proof.
let n := constr:(2) in
let p := constr:(S(n)) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_1);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_1;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjectures *)

Theorem true_1: forall x, E x -> N x.
Proof.
  do 2 intro.
apply (all_true_1 F_1);
 (trivial_in 0) ||
 (repeat constructor).
trivial.
Qed.


(* process the second strongly connected component *)

Local Definition type_LF_2 :=  T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_2 : type_LF_2:= (fun u1 => (O u1 -> N u1, ((model_nat u1):: nil))).

(* 2. build the list of Fs *)

Definition LF_2 := [F_2].

(* 3. build the main lemma *)
      
Lemma main_2 : forall F, In F LF_2 -> forall u1, (forall F', In F' LF_2 -> forall e1, less (snd (F' e1)) (snd (F u1)) -> fst (F' e1)) -> fst (F u1).
Proof.
intros F HF u1; case_In HF; intro Hind.
rename u1 into x.

simpl. intros.

(* instantiate x *) 
(* N(x) *) generalize Hind. inversion H; intros.

rename x0 into y. 
(* N(s(y) *) apply n2.

clear Hind H x H1.

(* instantiate y *) 

(* N(y) *) generalize Hind0. inversion H0; intros.

(* N(0) *) rewrite <- H. apply n1.

rename x into z.

(* N(s(z)) *) apply n2. clear Hind0 y H0 H1.

(* induction step for N(z) *)


(* N(z) *) apply (Hind1 F_2). 
simpl. left. trivial.
unfold snd. unfold F_2. rewrite_model. abstract solve_rpo_mul.

trivial. 
Qed.




Definition S_2 := fun f => exists F, In F LF_2 /\ exists e1, f = F e1.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_2: forall F, In F LF_2 -> forall u1: T, fst (F u1).
Proof.
let n := constr:(2) in
let p := constr:(S(n)) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_2);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_2;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjectures *)

Theorem true_2: forall x, O x -> N x.
Proof.
  do 2 intro.
apply (all_true_2 F_2);
 (trivial_in 0) ||
 (repeat constructor).
trivial.
Qed.

(* proving the root theorem, based on true_1 and true_2 *)

Theorem root_theorem : forall x, (E x \/ O x) -> N x.
Proof. 
intros. destruct H.

apply true_1. trivial.
apply true_2. trivial.
Qed.
Reset F_1.
End even_odd_02.


Section even_odd_03.

(* Coq certification for 03-even-or-odd-butterfly.tst *)

(* Initialisation of the proof process *)

Definition type_LF_1_2 :=  T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_1 : type_LF_1_2:= (fun u1 => (E u1 -> N u1, ((model_nat u1):: nil))).
Definition F_2 : type_LF_1_2:= (fun u1 => (O u1 -> N u1, ((model_nat u1):: nil))).

(* 2. build the list of Fs *)

Definition LF_1_2 := [F_1,F_2].

(* 3. build the main lemma *)
      
Lemma main_1_2 : forall F, In F LF_1_2 -> forall u1, (forall F', In F' LF_1_2 -> forall e1, less (snd (F' e1)) (snd (F u1)) -> fst (F' e1)) -> fst (F u1).
Proof.
intros F HF u1; case_In HF; intro Hind.
rename u1 into x.

simpl. intros.

(* instantiate x *) 
(* N(x) *) generalize Hind. inversion H; intros.

(* N(0) *) rewrite <- H0. apply n1.

rename x0 into y. 
(* N(s(y) *) apply n2.


clear Hind H x H1.

(* induction step for N(y) *)


(* N(z) *) apply (Hind0 F_2). 
simpl. right. left. trivial.
unfold snd. unfold F_2. unfold F_1. rewrite_model. abstract solve_rpo_mul.

trivial. 

simpl. intros.

rename u1 into x.

(* instantiate x *) 
(* N(x) *) generalize Hind. inversion H; intros.

rename x0 into y. 
(* N(s(y) *) apply n2.


clear Hind H x H1.

(* induction step for N(y) *)


(* N(z) *) apply (Hind0 F_1). 
simpl. left. trivial.
unfold snd. unfold F_2. unfold F_1. rewrite_model. abstract solve_rpo_mul.

trivial. 
Qed.

Definition S_1_2 := fun f => exists F, In F LF_1_2 /\ exists e1, f = F e1.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_1_2: forall F, In F LF_1_2 -> forall u1: T, fst (F u1).
Proof.
let n := constr:(2) in
let p := constr:(S(n)) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_1_2);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_1_2;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjectures *)

Theorem true_1: forall x, E x -> N x.
Proof.
  do 2 intro.
apply (all_true_1_2 F_1);
 (trivial_in 0) ||
 (repeat constructor).
trivial.
Qed.

Theorem true_2: forall x, O x -> N x.
Proof.
  do 2 intro.
apply (all_true_1_2 F_2);
 (trivial_in 1) ||
 (repeat constructor).
trivial.
Qed.

(* proving the root theorem, based on true_1 and true_2 *)

Theorem root_theorem : forall x, (E x \/ O x) -> N x.
Proof. 
intros. destruct H.

apply true_1. trivial.
apply true_2. trivial.
Qed.

End even_odd_03.

Section even_odd_04.

(* Coq certification for 04-n-is-even-or-odd.tst *)

(* Initialisation of the proof process *)

Definition type_LF_0 :=  T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_0 : type_LF_0:= (fun u1 => (N u1 -> (O u1 \/ E u1), ((model_nat u1):: nil))).

(* 2. build the list of Fs *)

Definition LF_0 := [F_0].

(* 3. build the main lemma *)
      
Lemma main_0 : forall F, In F LF_0 -> forall u1, (forall F', In F' LF_0 -> forall e1, less (snd (F' e1)) (snd (F u1)) -> fst (F' e1)) -> fst (F u1).
Proof.
intros F HF u1; case_In HF; intro Hind.
rename u1 into x.

simpl.

intros.

(* instantiate x *)
 generalize Hind.

(* N(x) *) inversion H. intros. rewrite <- H0. clear H0 Hind. 

(* O(0) \/ E(0) *) 

right. apply e1.

intros. rename x0 into y.

(* O(s(y)) \/ E(s(y)) *)

assert (O(y) \/ E(y)).

(* induction step for O(y) \/ E(y) *)


(* O(y) \/ E(y) *) apply (Hind0 F_0). 
simpl. left. trivial.
unfold snd. unfold F_0. rewrite_model. abstract solve_rpo_mul.

trivial. 

destruct H2.
right. apply e2. trivial.
left. apply o1. trivial.
Qed.

Definition S_0 := fun f => exists F, In F LF_0 /\ exists e1, f = F e1.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_0: forall F, In F LF_0 -> forall u1: T, fst (F u1).
Proof.
  let n := constr:(2) in
let p := constr:(S(n)) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_0);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_0;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjecture *)
      
Theorem true_0: forall u1:T, N u1 -> (O u1 \/ E u1).
Proof.
  do 1 intro.
apply (all_true_0 F_0).
 (trivial_in 0) ||
                (repeat constructor).
Qed.

End even_odd_04.
End evenodd.
