(* Coq certification for 17-hydra.tst from the E-Cyclist's distribution

Contact: sorin.stratulat@loria.fr

We use the certification methodology 
from our paper "Mechanically 
certifying formula-based {N}oetherian 
induction reasoning" 

*) 


(* import the syntactic ordering based on the Coccinelle library *)

Require Import ordering.

Section hydra.
(* T is the unique datatype. All the functions and inductive types are defined only on T. *)

Variable T: Set.
Variable zero: T.
Variable succ: T -> T.

Inductive p: T -> T -> Prop :=
  p1: p zero zero
| p2: p (succ zero) zero
| p3: forall x, p x (succ zero)
| p4: forall x y, p x y -> p (succ x) (succ (succ y))
| p5: forall y, p (succ y) y -> p zero (succ (succ y))
| p6: forall x, p (succ x) x -> p (succ (succ x)) zero
.

Inductive N: T -> Prop :=
  n1: N zero
| n2: forall x, N x -> N (succ x)
.
Variable model_nat : T -> term.

(** Initialisation of proof process *)
Axiom model_nat_0: model_nat zero = (Term id_0 nil).

Axiom  model_nat_succ: forall (u1: T), model_nat (succ u1) = (Term id_S ((model_nat u1)::nil)).

Hint Rewrite model_nat_0 model_nat_succ : model_nat.
Ltac rewrite_model := autorewrite with model_nat.



Definition type_LF_0_18 :=  T ->  T -> (Prop * (List.list term)).

(* 1. attach a measure to a root formula *)

Definition F_0 : type_LF_0_18:= (fun  u1 u2 => (N u1 -> N u2 -> p u1 u2, ((model_nat u1):: (model_nat u2):: nil))).

(* 2. build the list of Fs for each strongly connected component *)

Definition LF_0 := [F_0].


(* 3. build the main lemma *)
      
Lemma main_0 : forall F, In F LF_0 -> forall u1, forall u2, (forall F', In F' LF_0 -> forall e1, forall e2, less (snd (F' e1 e2)) (snd (F u1 u2)) -> fst (F' e1 e2)) -> fst (F u1 u2).
Proof.
intros F HF u1 u2; case_In HF; intro Hind.
rename u1 into x. rename u2 into y.
simpl.
intros. 
(* pxy *) generalize Hind.
inversion H0. rewrite <- H1. clear H1 H0 Hind. intros.
(* px0 *) generalize Hind.
inversion H. rewrite <- H0. clear H0 Hind. 
intros. 
(* p00 *) apply p1.
intros. clear Hind H1. rename x0 into x'.
(* psx'0 *) generalize Hind0. inversion H0. rewrite <- H1. clear Hind0 H1. intros. 
(* ps00 *) apply p2.
intros. clear Hind0 H2 x' y H H0. rename x0 into x''.
(* pssx''0 *) apply p6.

(* induction step  for psx''x'' *)

(* psx''x'' *) apply (Hind1 F_0).
simpl. left. trivial.
unfold snd. unfold F_0. rewrite_model. abstract solve_rpo_mul.

apply n2. trivial.
trivial. 

clear H2 Hind H0. intros. rename x0 into y'.
(* pxsy' *) generalize Hind.
inversion H1; intros. rewrite <- H0. 
(* pxs0 *) apply p3.

clear Hind y H1 H2 y'. rename x0 into y''.
(* pxssy'' *) generalize Hind0.
inversion H. rewrite <- H1. intros.
(* p0ssy'' *) apply p5. clear Hind0.

(* induction step for psy''y'' *)

(* psy''y'' *) apply (Hind1 F_0).
simpl. left. trivial.
unfold snd. unfold F_0. rewrite_model. abstract solve_rpo_mul.

apply n2. trivial.
trivial. 

intros. rename x0 into x'.
(* psx'ssy'' *) apply p4.

(* induction step for px'y''*)

(* px'y'' *) apply (Hind1 F_0).
simpl. left. trivial.
unfold snd. unfold F_0. rewrite_model. abstract solve_rpo_mul.
 trivial.
trivial. 

Qed. 

Definition S_0 := fun f => exists F, In F LF_0 /\ exists e1, exists e2, f = F e1 e2.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_0: forall F, In F LF_0 -> forall u1: T, forall u2: T, forall u3: T, fst (F u1  u2).
Proof.
  let n := constr:(2) in
let p := constr:(S(S(n))) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_0);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_0;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjectures *)

Theorem true_0: forall x y, N x -> N y -> p x y.
Proof.
  do 2 intro.
apply (all_true_0 F_0);
 (trivial_in 0) ||
 (repeat constructor).
trivial.
Qed.


End hydra.
