(* Coq certification for 05-p-and-q.tst from the Cyclist's distribution

Contact: sorin.stratulat@loria.fr

We use the certification methodology 
from our paper "Mechanically 
certifying formula-based {N}oetherian 
induction reasoning" 

*) 


(* import the syntactic ordering based on the Coccinelle library *)

Require Import ordering.

Section pq.
(* T is the unique datatype. All the functions and inductive types are defined only on T. *)

Variable T: Set.
Variable zero: T.
Variable succ: T -> T.

Inductive P: T -> Prop :=
  p1:  P zero
| p2: forall x, (P x /\ Q x (succ x)) -> P (succ x)
    with
Q: T -> T -> Prop :=
  q1: forall y, Q y zero
| q2: forall x y, (Q x y /\ P x) -> Q x (succ y).


Inductive N: T -> Prop :=
  n1: N zero
| n2: forall x, N x -> N (succ x)
.

(* the function translating Coq terms to Coccinelle terms *)

Variable model_nat : T -> term.
Axiom model_nat_0: model_nat zero = (Term id_0 nil).
Axiom  model_nat_succ: forall (u1: T), model_nat (succ u1) = (Term id_S ((model_nat u1)::nil)).

(* the equalities can be used for rewriting *)
Hint Rewrite model_nat_0 model_nat_succ : model_nat.
Ltac rewrite_model := autorewrite with model_nat.

(* Initialisation of the proof process. *)

Definition type_LF_0_4 :=  T ->  T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_0 : type_LF_0_4:= (fun  u1 u2 => (N u1 -> N u2 -> Q u1 u2, ((model_nat u1)::(model_nat u1)::(model_nat u2)::(model_nat u2):: nil))).

Definition F_4 : type_LF_0_4:= (fun  u1 u2 => (N u1 -> N u2 -> N (succ u2) -> Q u1 u2 /\ P u1, ((model_nat u1)::(model_nat u1)::(model_nat u2)::(model_nat u2):: (model_nat (succ u2)):: nil))).

(* 2. build the list of Fs *)

Definition LF_0_4 := [F_0,F_4].

(* 3. build the main lemma *)

Lemma main_0_4 : forall F, In F LF_0_4 -> forall u1, forall u2, (forall F', In F' LF_0_4 -> forall e1, forall e2, less (snd (F' e1 e2)) (snd (F u1 u2)) -> fst (F' e1 e2)) -> fst (F u1 u2).
Proof.
intros F HF u1 u2; case_In HF; intro Hind.
rename u1 into x. rename u2 into y.

simpl. intros.

(* instantiate y *)

(* Q(x,y) *) generalize Hind.
inversion H0. rewrite <- H1. clear Hind. intros.
(* Q(x,0) *) apply q1.

clear Hind. rename x0 into z. intros.
(* Q(x,s(z)) *) apply q2.


(* induction step for Q(x,z) /\ P(x) *)

(* Q(x,z) /\ P(x) *)  apply (Hind F_4).
simpl. right. left. trivial.
unfold snd. unfold F_4. unfold F_0. rewrite_model. abstract solve_rpo_mul.
trivial.
trivial. apply n2. trivial. 

simpl. intros. rename u1 into x. rename u2 into z.
split. 

(* induction step for Q(x,y) *)

(* Q(x,y) *)  apply (Hind F_0).
simpl. left. trivial.
unfold snd. unfold F_4. unfold F_0. rewrite_model. abstract solve_rpo_mul.
trivial.
trivial. 

clear H1 H0.

(* instantiate x *)

(* P(x) *) generalize Hind.
inversion H. rewrite <- H0. intros.
(* P(0) *) apply p1.

intros. rename x0 into y. 
(* P(s(y)) *) apply p2.

clear Hind H1 x H.
assert (Q y y /\ P y).

(* induction step for Q(y,y) /\ P(y) *)

(* Q(y,y) /\ P(y) *) apply (Hind0 F_4). 
simpl. right. left. trivial.
unfold snd. unfold F_4. unfold F_0. rewrite_model. abstract solve_rpo_mul.

trivial. trivial. apply n2. trivial.

destruct H. split. trivial.

apply q2. split; trivial.
Qed.

Definition S_0_4 := fun f => exists F, In F LF_0_4 /\ exists e1, exists e2, f = F e1 e2.

(* 4. prove all formulas in LF_0_4, using the well-founded induction principle *)

Theorem all_true_0_4: forall F, In F LF_0_4 -> forall u1: T, forall u2: T, fst (F u1  u2).
Proof.
let n := constr:(2) in
let p := constr:(S(S(n))) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_0_4);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_0_4;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjecture *)

Theorem true_0: forall x y, N x -> N y -> Q x y.
Proof.

(* the pre-proof without induction reasoning *)

(* intros x y H H0. *)
(* (* instantiate y from Q(x,y) *) *)
(* inversion H0. rewrite <- H1. *)
(* (* Q(x,0) *) apply q1. *)
(* rename x0 into z.  *)
(* (* Q(x,s(z)) *) apply q2. split.  *)
(* (* the proof of Q(x,z) requires induction reasoning *) *)
(* Focus 2. *)

(* (* instantiate x from  P(x) *) *)
(* inversion H. rewrite <- H3. *)
(* (* P(0) *) apply p1. *)
(* clear y H0 H2. rename x0 into y.  *)
(* (* P(s(y)) *) apply p2. *)
(* assert (Q y y /\ P y). *)
(* (* the proof of Q(y,y) /\ P(y) requires induction reasoning *) *)
(* Focus 2. *)

(* (* show that Q y y /\ P y and P y /\ Q y (succ y) are equivalent *) *)
(* destruct H0. split; trivial. apply q2. split; trivial. *)
(* Admitted. *)

  do 2 intro.
apply (all_true_0_4 F_0);
 (trivial_in 0) ||
 (repeat constructor).
Qed.



Theorem true_4: forall x y,  N x -> N y -> N (succ y) -> Q x y /\ P x.
Proof.
  do 2 intro.
apply (all_true_0_4 F_4);
 (trivial_in 1) ||
 (repeat constructor).
Qed.

End pq.
