
(* Coq certification for 12-r-example.tst from the Cyclist's distribution

Contact: sorin.stratulat@loria.fr

We use the certification methodology 
from our paper "Mechanically 
certifying formula-based {N}oetherian 
induction reasoning" 

*) 

(* import the syntactic ordering based on the Coccinelle library *)

Require Import ordering.

Section r.
(* T is the unique datatype. All the functions and inductive types are defined only on T. *)

Variable T: Set.
Variable zero: T.
Variable succ: T -> T.


Inductive R: T -> T -> Prop :=
  r1: forall y, R zero y
| r2: forall x, R x zero -> R (succ x) zero
| r3: forall x y, R (succ (succ zero)) y -> R (succ x) (succ y).

Inductive N: T -> Prop :=
  n1 : N zero
| n2 : forall x, N x -> N (succ x).


(* the translation function to Coccinelle terms *)

Variable model_nat : T -> term.
Axiom model_nat_0: model_nat zero = (Term id_0 nil).
Axiom  model_nat_succ: forall (u1: T), model_nat (succ u1) = (Term id_S ((model_nat u1)::nil)).

Hint Rewrite model_nat_0 model_nat_succ : model_nat.
Ltac rewrite_model := autorewrite with model_nat.

(* Initialisation of the proof process. Process the first strongly connected component *)


Definition type_LF_18 :=  T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_18 : type_LF_18:= (fun  u1 => (N u1 -> N (succ (succ u1)) -> N (succ u1) -> N zero -> R (succ u1) zero, ((model_nat u1):: nil))).

(* 2. build the list of Fs for each strongly connected component *)

Definition LF_18 := [F_18].

(* 3. build the main lemma *)
      
Lemma main_18 : forall F, In F LF_18 -> forall u1, (forall F', In F' LF_18 -> forall e1, less (snd (F' e1)) (snd (F u1)) -> fst (F' e1)) -> fst (F u1).
Proof.
intros F HF u1; case_In HF; intro Hind.
rename u1 into x.

simpl.

intros.
(* instantiate x *)

 inversion H.

(* base case: x = zero *)
rewrite <- H3. apply r2.
apply r1.

(* step case *)
apply r2.

(* found r (succ x) zero  and apply the IH *)
apply (Hind F_18); trivial.
simpl. left. trivial.

(* check the ordering constraints *)
rewrite <- H4. unfold snd. unfold F_18. rewrite_model. abstract solve_rpo_mul.

(* validate the remaining conditions *)

rewrite <- H4 in H1. trivial.
rewrite H4. trivial.
Qed.

Definition S_18 := fun f => exists F, In F LF_18 /\ exists e1, f = F e1.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_18: forall F, In F LF_18 -> forall u1: T, fst (F u1).
Proof.
  let n := constr:(2) in
let p := constr:(S(n)) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_18);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_18;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjectures *)
      
Theorem true_18: forall x:T, N x -> N (succ (succ x)) -> N (succ x) -> N zero -> R (succ x) zero.
Proof.
  do 1 intro.
apply (all_true_18 F_18).
 (trivial_in 0) ||
                (repeat constructor).
Qed.

(* process the second strongly connected component *)

Definition type_LF_0 :=  T -> T -> (Prop * (List.list term)).

(* 1. attach a measure to each root formula *)

Definition F_0 : type_LF_0:= (fun  u1 u2 => (N u1 -> N u2 -> R u1 u2, ((model_nat u2):: nil))).

(* 2. build the list of Fs for each strongly connected component *)

Definition LF_0 := [F_0].

(* 3. build the main lemma *)
      
Lemma main_0 : forall F, In F LF_0 -> forall u1, forall u2, (forall F', In F' LF_0 -> forall e1, forall e2, less (snd (F' e1 e2)) (snd (F u1 u2)) -> fst (F' e1 e2)) -> fst (F u1 u2).
Proof.
intros F HF u1 u2; case_In HF; intro Hind.
rename u1 into x. rename u2 into y.

simpl. intros.

(* instantiate x *) 
inversion H. rewrite <- H1. 

(* case x = zero *)
apply r1.

(* case x = succ x0 *)

(* instantiate x0 *)
inversion H1.

(* case x0 = zero *)

(* instantiate y *)
inversion H0. rewrite <- H4. 

(* case y = zero *)
  apply r2.
  rewrite <- H3. apply r1.

  (* case y = succ y0 *)
  apply r3.
apply (Hind F_0); trivial.
simpl. left. trivial.
 unfold snd. unfold F_0. rewrite <- H5. rewrite_model. abstract solve_rpo_mul.

 apply n2. apply n2. apply n1.

 (* instantiate y *)
 inversion H0.

 (* case y = zero *)
  rewrite <- H5. apply r2.

 (* instantiate x1 *)
  inversion H3.

  (* case x1 = zero *)
  rewrite <- H6. apply r2.
  apply r1.

  (* case x1 = succ x2 *)
  apply r2.
  apply true_18.

  (* get rid off the conditions *)

  trivial.
  rewrite  H7. rewrite H4. trivial.
  rewrite H7. trivial.
  apply n1.

  (* case y = succ x2 *)
  apply r3.

  (* instantiate x2 *)
  inversion H5.

  (* case x2 = zero *)
  apply (Hind F_0).
  simpl. left. trivial.
  unfold snd. unfold F_0. rewrite <- H6. rewrite_model. abstract solve_rpo_mul.

 (* get rid off the conditions *)
 apply n2. apply n2. rewrite <- H7. apply n1.
 rewrite <- H7. apply n1.

 (* case x2 = succ x3 *)
 apply (Hind F_0).
simpl. left. trivial.
 unfold snd. unfold F_0. rewrite <- H6. rewrite <- H8. rewrite_model. abstract solve_rpo_mul.

 (* get rid off the conditions *)
 apply n2. apply n2. apply n1.
 rewrite H8. trivial.
 

Qed. 

Definition S_0 := fun f => exists F, In F LF_0 /\ exists e1, exists e2, f = F e1 e2.

(* 4. check all formulas in LF_0, using the well-founded induction principle *)

Theorem all_true_0: forall F, In F LF_0 -> forall u1: T, forall u2: T, fst (F u1  u2).
Proof.
  let n := constr:(2) in
let p := constr:(S(S(n))) in
intros;
let G := fresh "G" in
let x := fresh "x" in
apply wf_subset with (R:=@snd_rpo_mul Pr Prop max_size) (S:=S_0);
[(* 1 *) apply wf_snd_rpo_mul, prec_wf
|(* 2 *) idtac
|(* 3 *) eexists; split; [ eassumption | idtac]; do_nat n ltac:(eexists); reflexivity
];

intros x G;
do_nat p ltac:(elim G; intro; clear G; intro G);
rewrite G in * |- *; clear G; clear x;
intro G;
apply main_0;
 [assumption | idtac];
intros;
apply G;
 [ idtac | assumption ];
eexists; split;
 [idtac | do_nat n ltac:(eexists); reflexivity];
assumption.
Qed.

(* 5. prove the root conjectures *)

Theorem true_0: forall x y, N x -> N y -> R x y.
Proof.
  do 2 intro.
apply (all_true_0 F_0);
 (trivial_in 0) ||
 (repeat constructor).
Qed.

End r.
